"""
Plot overall classification performance metrics.
"""

import os
import sys
import numpy as np
import itertools
from distutils.spawn import find_executable
from sklearn.metrics import roc_curve, auc
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import average_precision_score
from scipy import interp

try:
    import matplotlib
    import matplotlib.pyplot as plt

    # Check if latex is installed
    if find_executable('latex'):
        plt.rcParams['text.usetex'] = True
    plt.rcParams['font.serif'] = ['Computer Modern Roman'] + plt.rcParams['font.serif']

    font = {'family': 'normal',
            'size': 34}

    matplotlib.rc('font', **font)
except ImportError:
    print("Warning: You will need to install matplotlib if you want to plot any metric")

COLORS = ['tab:green', 'tab:orange', 'tab:blue', 'tab:red', 'tab:purple', 'tab:brown', '#aaffc3', 'tab:olive',
          'tab:cyan', '#FF1493', 'navy', 'tab:pink', 'lightcoral', '#228B22', '#aa6e28', '#FFA07A']


def plasticc_log_loss(y_true, y_pred, relative_class_weights=None):
    """
    Implementation of weighted log loss used for the Kaggle challenge
    """

    if np.nonzero(y_true[:, 0])[0].size == 0:
        start_index = 1
    else:
        start_index = 0
    print(start_index)

    predictions = y_pred.copy()

    # sanitize predictions
    epsilon = sys.float_info.epsilon  # this is machine dependent but essentially prevents log(0)
    predictions = np.clip(predictions, epsilon, 1.0 - epsilon)
    predictions = predictions / np.sum(predictions, axis=1)[:, np.newaxis]

    predictions = np.log(predictions)
    # multiplying the arrays is equivalent to a truth mask as y_true only contains zeros and ones
    class_logloss = []
    for i in range(start_index, predictions.shape[1]):
        # average column wise log loss with truth mask applied
        result = np.average(predictions[:, i][y_true[:, i] == 1])
        class_logloss.append(result)
    return -1 * np.average(class_logloss, weights=relative_class_weights[start_index:])


def compute_precision_recall(classes, y_test, y_pred_prob, name='', fig_dir='.', title=None):
    """
    Plot Precision-Recall curves.
    """

    if np.nonzero(y_test[:, 0])[0].size == 0:
        start_index = 1
    else:
        start_index = 0

    nclasses = len(classes)
    # For each class
    precision = dict()
    recall = dict()
    save_auc = dict()
    average_precision = dict()
    for i in range(start_index, nclasses):
        precision[i], recall[i], _ = precision_recall_curve(y_test[:, i], y_pred_prob[:, i])
        average_precision[i] = average_precision_score(y_test[:, i], y_pred_prob[:, i])
        save_auc[classes[i]] = average_precision[i]

    # A "micro-average": quantifying score on all classes jointly
    precision["micro"], recall["micro"], _ = precision_recall_curve(y_test.ravel(), y_pred_prob.ravel())
    average_precision["micro"] = average_precision_score(y_test, y_pred_prob,
                                                         average="micro")
    save_auc[classes[i]] = average_precision["micro"]
    print('Average precision score, micro-averaged over all classes: {0:0.2f}'
          .format(average_precision["micro"]))

    plt.figure(figsize=(12, 16))
    f_scores = np.linspace(0.2, 0.8, num=4)
    lines = []
    labels = []
    for f_score in f_scores:
        x = np.linspace(0.01, 1)
        y = f_score * x / (2 * x - f_score)
        # l, = plt.plot(x[y >= 0], y[y >= 0], color='gray', alpha=0.2)
        # plt.annotate('f1={0:0.1f}'.format(f_score), xy=(0.9, y[45] + 0.02))

    # lines.append(l)
    # labels.append('iso-f1 curves')
    l, = plt.plot(recall["micro"], precision["micro"], color='navy', linestyle=':', lw=2)
    lines.append(l)
    labels.append('micro-average ({0:0.2f})'
                  ''.format(average_precision["micro"]))

    for i in range(start_index, nclasses):
        l, = plt.plot(recall[i], precision[i], color=COLORS[i], lw=2)
        lines.append(l)
        labels.append('{0} ({1:0.2f})'
                      ''.format(classes[i], average_precision[i]))

    fig = plt.gcf()
    fig.subplots_adjust(bottom=0.25)
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    if title is not None:
        plt.title(title, fontsize=34)
    plt.legend(lines, labels, loc=(0.1, -.6), fontsize=24, frameon=True, ncol=2)
    plt.tight_layout()
    figname = os.path.join(fig_dir, 'precision_%s.pdf' % name)
    plt.savefig(figname)
    figname = os.path.join(fig_dir, 'precision_%s.png' % name)
    plt.savefig(figname)
    plt.close()

    return figname, save_auc


def compute_multiclass_roc_auc(classes, y_test, y_pred_prob, name='', fig_dir='.', title=None, logyscale=False):
    """
    Plot multiclass Receiver Operating Characteristic curves.
    """

    if np.nonzero(y_test[:, 0])[0].size == 0:
        start_index = 1
    else:
        start_index = 0

    nclasses = len(classes)

    # Compute ROC curve and ROC area for each class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    save_auc = dict()
    for i in range(start_index, nclasses):
        fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_pred_prob[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])
        save_auc[classes[i]] = roc_auc[i]

    # Compute micro-average ROC curve and ROC area
    fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), y_pred_prob.ravel())
    roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])
    save_auc['micro'] = roc_auc['micro']

    # Compute macro-average ROC curve and ROC area

    # First aggregate all false positive rates
    all_fpr = np.unique(np.concatenate([fpr[i] for i in range(start_index, nclasses)]))

    # Then interpolate all ROC curves at this points
    mean_tpr = np.zeros_like(all_fpr)
    for i in range(start_index, nclasses):
        mean_tpr += interp(all_fpr, fpr[i], tpr[i])

    # Finally average it and compute AUC
    mean_tpr /= len(classes[start_index:])

    fpr["macro"] = all_fpr
    tpr["macro"] = mean_tpr
    roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])
    save_auc['macro'] = roc_auc['macro']

    # Plot all ROC curves
    fig = plt.figure(figsize=(13, 12))
    plt.plot(fpr["micro"], tpr["micro"],
             label='micro-average ({0:0.2f})'
                   ''.format(roc_auc["micro"]),
             color='navy', linestyle=':', linewidth=6)

    plt.plot(fpr["macro"], tpr["macro"],
             label='macro-average ({0:0.2f})'
                   ''.format(roc_auc["macro"]),
             color='deeppink', linestyle=':', linewidth=6)

    lw = 2
    # colors = cycle(['aqua', 'darkorange', 'cornflowerblue'])
    for i in range(start_index, nclasses):
        plt.plot(fpr[i], tpr[i], lw=lw, color=COLORS[i],
                 label='{0} ({1:0.2f})'
                       ''.format(classes[i], roc_auc[i]))

    # # plt.plot([0, 1], [0, 1], 'k--', lw=lw)
    # plt.xlim([0.0, 1.0])
    if logyscale:
        plt.yscale("log")
    else:
        pass  # plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    if title is not None:
        plt.title(title, fontsize=34)  # plt.title(title, fontsize=70, fontweight="bold", y=1.02) # was size 34
    plt.legend(loc="lower right", frameon=True, fontsize=26)
    plt.tight_layout()
    figname = os.path.join(fig_dir, 'roc_%s_mask_rare.pdf' % name)
    plt.savefig(figname)
    figname = os.path.join(fig_dir, 'roc_%s.png' % name)
    plt.savefig(figname)

    return figname, save_auc


def plot_confusion_matrix(cm, classes, normalize=False, title=None, cmap=plt.cm.RdBu, fig_dir='.', name='',
                          combine_kfolds=False, show_uncertainties=False):
    """
    Plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """

    if combine_kfolds:
        uncertainties = np.std(cm, axis=0)
        cm = np.sum(cm, axis=0)

    if normalize:
        if combine_kfolds:
            uncertainties = uncertainties.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print('Confusion matrix, without normalization')

    print(cm)
    # Multiply off diagonal by -1
    off_diag = ~np.eye(cm.shape[0], dtype=bool)
    cm[off_diag] *= -1
    np.savetxt(os.path.join(fig_dir, 'confusion_matrix_%s.csv' % name), cm)
    print(cm)

    cms = [cm]
    deleterows = [False]
    if np.all(np.isnan(cm[0])):
        cmDelete = np.delete(cm, 0, 0)
        cms.append(cmDelete)
        deleterows.append(True)

    for cm, deleterow in zip(cms, deleterows):
        fig = plt.figure(figsize=(15, 12))
        plt.imshow(cm, interpolation='nearest', cmap=cmap, vmin=-1, vmax=1)
        # plt.title(title)
        cb = plt.colorbar()
        cb.ax.set_yticklabels(cb.ax.get_yticklabels(), fontsize=27)
        tick_marks = np.arange(len(classes))
        plt.xticks(tick_marks, classes, rotation=90, fontsize=27)
        if deleterow:
            plt.yticks(tick_marks[:-1], classes[1:], fontsize=27)
        else:
            plt.yticks(tick_marks, classes, fontsize=27)

        fmt = '.2f' if normalize else 'd'
        thresh = 0.5  # cm.max() / 2.
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            value = format(abs(cm[i, j]), fmt)
            if combine_kfolds and show_uncertainties:
                unc = format(uncertainties[i, j], fmt)
                cell_text = r"{} $\pm$ {}".format(value, unc)
            else:
                cell_text = value
            if cell_text == 'nan':
                cell_text = '-'
            plt.text(j, i, cell_text, horizontalalignment="center",
                     color="white" if abs(cm[i, j]) > thresh else "black", fontsize=26)

        if title is not None:
            plt.title(title, fontsize=34)  # plt.title(title, fontsize=70, fontweight="bold", y=1.02) # was size 33
        plt.ylabel('True label')
        plt.xlabel('Predicted label')
        figname_pdf = os.path.join(fig_dir, 'confusion_matrix_%s_mask_rare.pdf' % name)
        plt.savefig(figname_pdf, bbox_inches="tight")
        if not deleterow:
            figname_png = os.path.join(fig_dir, 'confusion_matrix_%s.png' % name)
            plt.savefig(figname_png, bbox_inches="tight")

    return figname_png



# -*- coding: UTF-8 -*-
"""
Get PLASTICC data from SQL database
"""
import sys
import os
import numpy as np
import warnings
import argparse
import pandas as pd
import astropy.table as at
import astropy.io.fits as afits
from collections import OrderedDict
from . import database
from astrorapid import helpers

ROOT_DIR = os.getenv('PLASTICC_DIR')
DATA_DIR = os.path.join(ROOT_DIR, 'plasticc_data')


def parse_getdata_options(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    def str2bool(v):
        return v.lower() in ("yes", "true", "t", "1")

    def_sql_wildcard = '%'
    parser = argparse.ArgumentParser(description="Get options to the GetData structure")
    group = parser.add_mutually_exclusive_group(required=False)
    parser.register('type', 'bool', str2bool)
    parser.add_argument('--data_release', required=True, help='PLAsTiCC data release index table to process')
    field_choices = ('WFD', 'DDF', '%')
    parser.add_argument('--field', required=False, default='DDF', type=str.upper, choices=field_choices, \
                        help='PLAsTiCC field to process')

    type_mapping = GetData.get_sntypes()
    inverse_mapping = {v: k for k, v in type_mapping.items()}

    model_choices = list(type_mapping.keys()).append('%')
    model_name_choices = list(type_mapping.values()).append('%')
    group.add_argument('--model', required=False, action="store", default=def_sql_wildcard, choices=model_choices, \
                       help='PLAsTiCC model to process')
    group.add_argument('--model_name', required=False, action="store", default=def_sql_wildcard,
                       choices=model_name_choices, \
                       help='PLAsTiCC model name to process')
    parser.add_argument('--base', required=False, default=def_sql_wildcard,
                        help='PLAsTiCC model base filename (probably not a good idea to touch this)')
    parser.add_argument('--snid', required=False, default=def_sql_wildcard,
                        help='PLAsTiCC object ID number (useful for debugging/testing)')
    parser.add_argument('--limit', required=False, type=int, default=None,
                        help='Limit the number of returned results from the MySQL index')
    parser.add_argument('--shuffle', required=False, type="bool", default="False",
                        help='Shuffle the returned results from the MySQL index')
    parser.add_argument('--sort', required=False, type="bool", default="True",
                        help='Sort the returned results from the MySQL index')
    parser.add_argument('--survey', required=False, default="LSST", 
                        help="Specify the survey to process")
    parser.add_argument('--offset', required=False, default=None, type=int,
                        help='Return the MySQL results AFTER offset rows')
    parser.add_argument('--extrasql', required=False, default=None,
                        help='Extra SQL for the selection function - enter as quoted string - used as is')
    args = parser.parse_args(args=argv)

    out = vars(args)
    model_name = out.pop('model_name')
    model = out.pop('model')
    if model_name == '%':
        if model == '%':
            out['model'] = '%'
        else:
            out['model'] = model
    else:
        this_model = inverse_mapping.get(model_name)
        if model == '%':
            out['model'] = this_model
        else:
            if this_model == model:
                out['model'] = this_model
            else:
                out['model'] = model
    return out


class GetData(object):
    """
    Class to access the ANTARES parsed PLaSTiCC index and light curve data
    """

    def __init__(self, data_release):
        self.data_release = "release_{}".format(data_release)
        self.phot_fields = ['MJD', 'FLT', 'FLUXCAL', 'FLUXCALERR', 'ZEROPT', 'PHOTFLAG']
        self.phot_fields_dtypes = {'FLT': np.str_, 'PHOTFLAG': np.int_}

    def get_phot_fields(self):
        """
        list of the photometry column names and a dictionary of NON-FLOAT
        columns

        For the default columns, this is only FLT
        """
        return list(self.phot_fields), dict(self.phot_fields_dtypes)

    def set_phot_fields(self, fields, dtypes):
        """
        set the list of photometry column fields to retrieve and a dictionary
        of data types for NON-FLOAT columns hashed by column name in the PHOT.FITS

        i.e. if your column is a float, don't even bother listing it

        Can be used to retrieve custom columns from the PHOT.FITS
        Kinda kludgey - se with caution
        """
        self.phot_fields = list(fields)
        self.phot_fields_dtypes = dict(dtypes)

    def get_object_ids(self):
        """ Get list of all object ids """
        obj_ids = database.exec_sql_query("SELECT objid FROM {0};".format(self.data_release))
        return obj_ids

    def get_column_for_sntype(self, column_name, sntype, field='%'):
        """ Get an sql column for a particular sntype class

        Parameters
        ----------
        column_name : str
            column name. E.g. column_name='peakmjd'
        sntype : int
            sntype number. E.g. sntype=4
        field : str, optional
            The field name. E.g. field='DDF' or field='WFD'. The default is '%' indicating that all fields will be included.

        Return
        -------
        column_out: list
            A list containing all the entire column for a particular sntype class
        """
        try:
            column_out = database.exec_sql_query(
                "SELECT {0} FROM {1} WHERE objid LIKE '{2}%' AND sntype={3};".format(column_name, self.data_release,
                                                                                     field, sntype))
            column_out = np.array(column_out)[:, 0]
        except IndexError:
            print("No data in the database satisfy the given arguments. field: {}, sntype: {}".format(field, sntype))
            return []
        return column_out

    def get_photfile_for_objid(self, objid):
        """
        Returns the phot file for the object ID
        """
        field, model, base, snid = objid.split('_')
        if field == 'IDEAL':
            filename = "{0}_MODEL{1}/{0}_{2}_PHOT.FITS".format(field, model, base)
        elif field == 'MSIP':
            filename = "ZTF_{0}_MODEL{1}/ZTF_{0}_{2}_PHOT.FITS".format(field, model, base)
        else:
            filename = "LSST_{0}_MODEL{1}/LSST_{0}_{2}_PHOT.FITS".format(field, model, base)
        phot_file = os.path.join(DATA_DIR, self.data_release.replace('release_', ''), filename)
        if not os.path.exists(phot_file):
            phot_file = phot_file + '.gz'
        return phot_file



    def get_light_curve(self, objid, ptrobs_min, ptrobs_max, standard_zpt=27.5):
        """ Get lightcurve from fits file

        Parameters
        ----------
        objid : str
            The object ID. E.g. objid='DDF_04_NONIa-0004_87287'
        ptrobs_min : int
            Min index of object in _PHOT.FITS.
        ptrobs_max : int
            Max index of object in _PHOT.FITS.

        Return
        -------
        phot_out: pandas DataFrame
            A DataFrame containing the MJD, FLT, FLUXCAL, FLUXCALERR, ZEROPT seperated by each filter.
            E.g. Access the magnitude in the z filter with phot_out['z']['MAG'].
        """
        phot_file = self.get_photfile_for_objid(objid)

        try:
            phot_HDU = afits.open(phot_file, memmap=True)
        except Exception as e:
            message = f'Could not open photometry file {phot_file}'
            raise RuntimeError(message)

        phot_data = phot_HDU[1].data[ptrobs_min - 1:ptrobs_max]

        phot_dict = OrderedDict()
        filters = list(set(phot_data['FLT']))  # e.g. ['i', 'r', 'Y', 'u', 'g', 'z']
        dtypes = dict(self.phot_fields_dtypes)
        for f in filters:
            fIndexes = np.where(phot_data['FLT'] == f)[0]
            phot_dict[f] = OrderedDict()
            for pfield in self.phot_fields:
                if pfield == 'ZEROPT':
                    phot_dict[f][pfield] = np.repeat(standard_zpt, len(fIndexes))
                elif pfield == 'FLT':
                    true_zpt = phot_data['ZEROPT'][fIndexes]
                    nobs = len(true_zpt)
                    phot_dict[f][pfield] = np.repeat(f.strip(), nobs)
                else:
                    phot_dict[f][pfield] = phot_data[pfield][fIndexes]

                if not pfield in dtypes:
                    dtypes[pfield] = np.float64

        phot_out = pd.DataFrame(phot_dict)
        phot_HDU.close()
        del phot_HDU[1].data
        return phot_out

    def get_light_curve_array(self, objid, ptrobs_min, ptrobs_max, standard_zpt=27.5):
        """ Get lightcurve from fits file as an array - avoid some Pandas overhead

        Parameters
        ----------
        objid : str
            The object ID. E.g. objid='DDF_04_NONIa-0004_87287'
        ptrobs_min : int
            Min index of object in _PHOT.FITS.
        ptrobs_max : int
            Max index of object in _PHOT.FITS.

        Return
        -------
        phot_out: pandas DataFrame
            A DataFrame containing the MJD, FLT, FLUXCAL, FLUXCALERR, ZEROPT seperated by each filter.
            E.g. Access the magnitude in the z filter with phot_out['z']['MAG'].
        """
        phot_file = self.get_photfile_for_objid(objid)

        try:
            phot_HDU = afits.open(phot_file, memmap=True)
        except Exception as e:
            message = f'Could not open photometry file {phot_file}'
            raise RuntimeError(message)

        phot_data = phot_HDU[1].data[ptrobs_min - 1:ptrobs_max]
        phot_data = at.Table(phot_data)
        return phot_data


    @staticmethod
    def convert_pandas_lc_to_recarray_lc(phot, passbands=('u', 'g', 'r', 'i', 'z', 'Y')):
        """
        ANTARES_object not Pandas format broken up by passband
        TODO: This is ugly - just have an option for get_lcs_data to return one or the other
        """
        pbs = passbands
        # name mapping for the defaults in phot_fields
        # any other column is going to become just lowercase it's current name
        name_map = {'FLUXCAL': 'flux', 'FLUXCALERR': 'dflux', 'ZEROPT': 'zpt',\
                        'FLT': 'pb', 'MJD': 'mjd', 'PHOTFLAG': 'photflag'}

        out = None
        out_names = None

        for this_pb in phot:
            # do we know what this passband is
            # this is just a sanity test in case we accidentally retrieve dummy entries with passband = -9
            if this_pb not in pbs:
                continue

            this_pb_lc = phot.get(this_pb)
            if this_pb_lc is None:
                continue

            if out is None:
                out = np.rec.fromarrays(list(this_pb_lc))
                if out_names is None:
                    out_names = list(this_pb_lc.axes[0])
                    out_names = [name_map.get(x, x.lower()) for x in out_names]
            else:
                temp = np.rec.fromarrays(list(this_pb_lc))
                out = np.concatenate((out, temp), axis=-1)

        out.dtype.names = out_names
        return out

    @staticmethod
    def get_sntypes():
        return helpers.get_sntypes()

    @staticmethod
    def aggregate_sntypes(reverse=False):
        return helpers.aggregate_sntypes(reverse=reverse)

    def get_avail_sntypes(self):
        """ Returns a list of the different transient classes in the database. """
        sntypes = database.exec_sql_query("SELECT DISTINCT sntype FROM {};".format(self.data_release))
        sntypes_map = self.get_sntypes()
        return sorted([sntype[0] for sntype in sntypes]), sntypes_map

    def get_lcs_headers(self, columns=None, field='%', model='%', base='%', snid='%', extrasql='', survey='LSST',
                        get_num_lightcurves=False, limit=None, shuffle=False, sort=True, offset=0, big=False):
        """ Gets the header data given specific conditions.

        Parameters
        ----------
        columns : list
            A list of strings of the names of the columns you want to retrieve from the database.
            You must at least include ['objid', 'ptrobs_min', 'ptrobs_max'] at the beginning of the input list.
            E.g. columns=['objid', 'ptrobs_min', 'ptrobs_max', 'sntype', 'peakmjd'].
        field : str, optional
            The field name. E.g. field='DDF' or field='WFD'. The default is '%' indicating that all fields will be included.
        model : str, optional
            The model number. E.g. model='04'. The default is '%' indicating that all model numbers will be included.
        base : str, optional
            The base name. E.g. base='NONIa'. The default is '%' indicating that all base names will be included.
        snid : str, optional
            The transient id. E.g. snid='87287'. The default is '%' indicating that all snids will be included.
        get_num_lightcurves : boolean, optional
            If this is True, then the return value is just a single iteration generator stating the number of
            light curves that satisfied the given conditions.
        limit : int, optional 
            Limit the results to this number (> 0)
        shuffle : bool, optional
            Randomize the order of the results - not allowed with `sort`
        sort : bool, optional
            Order the results by objid - overrides `shuffle` if both are set
        offset : int, optional
            Start returning MySQL results from this row number offset
        big : bool, optional
            If True, use a generator to retrieve results - cannot be used with
            get_num_lightcurves since generators have no length
        Return
        -------
        result: tuple or int
        """
        if columns is None:
            columns = ['objid', 'ptrobs_min', 'ptrobs_max']

        if get_num_lightcurves:
            columns = ['COUNT(objid)', ]
            big = False

        try:
            limit = int(limit)
            if limit <= 0:
                raise RuntimeError('prat')
        except Exception as e:
            limit = None

        try:
            offset = int(offset)
            if offset <= 0:
                raise RuntimeError('prat')
        except Exception as e:
            offset = None

        if limit is not None and shuffle is False and sort is False:
            sort = True

        extrasql_command = '' if extrasql is None else extrasql
        limit_command = '' if limit is None else " LIMIT {}".format(limit)
        offset_command = '' if offset is None else " OFFSET {}".format(offset)
        if model != '%':
            model = "{:02n}".format(int(model))

        if sort is True and shuffle is True:
            message = 'Cannot sort and shuffle at the same time! That makes no sense!'
            shuffle = False
            warnings.warn(message, RuntimeWarning)

        shuffle_command = '' if shuffle is False else " ORDER BY RAND()"
        sort_command = '' if sort is False else ' ORDER BY objid'
        extra_command = ''.join([extrasql_command, sort_command, shuffle_command, limit_command, offset_command])

        query = "SELECT {} FROM {} WHERE objid LIKE '{}_{}_{}_{}' {};".format(', '.join(columns), \
                                                                              self.data_release, field, model, base,
                                                                              snid, extra_command)
        if big:
            for result in database.exec_big_sql_query(query):
                yield result
        else:
            header = database.exec_sql_query(query)
            if get_num_lightcurves:
                num_lightcurves = int(header[0][0])
                yield num_lightcurves
            else:
                num_lightcurves = len(header)

            if num_lightcurves > 0:
                for result in header:
                    yield result
            else:
                print("No light curves in the database satisfy the given arguments. "
                      "field: {}, model: {}, base: {}, snid: {}".format(field, model, base, snid))
                return

    def get_lcs_data(self, columns=None, field='%', model='%', base='%', snid='%', survey='LSST',\
                     limit=None, shuffle=False, sort=True, offset=0, big=False, extrasql=''):
        """ Gets the light curve and header data given specific conditions. Returns a generator of LC info.

        Parameters
        ----------
        columns : list
            A list of strings of the names of the columns you want to retrieve from the database.
            You must at least include ['objid', 'ptrobs_min', 'ptrobs_max'] at the beginning of the input list.
            E.g. columns=['objid', 'ptrobs_min', 'ptrobs_max', 'peakmjd'].
        field : str, optional
            The field name. E.g. field='DDF' or field='WFD'. The default is '%' indicating that all fields will be included.
        model : str, optional
            The model number. E.g. model='04'. The default is '%' indicating that all model numbers will be included.
        base : str, optional
            The base name. E.g. base='NONIa'. The default is '%' indicating that all base names will be included.
        snid : str, optional
            The transient id. E.g. snid='87287'. The default is '%' indicating that all snids will be included.
        limit : int, optional 
            Limit the results to this number (> 0)
        shuffle : bool, optional
            Randomize the order of the results - not allowed with `sort`
        sort : bool, optional
            Order the results by objid - overrides `shuffle` if both are set
        offset : int, optional
            Start returning MySQL results from this row number offset (> 0)
        big : bool, optional
            If True, use a generator to retrieve results - cannot be used with
            get_num_lightcurves since generators have no length
            

        Return
        -------
        result: tuple
            A generator tuple containing (objid, ptrobs_min, ptrobs_max, mwebv, mwebv_err, z, zerr, peak_mjd)
        phot_data : pandas DataFrame
            A generator containing a DataFrame with the MJD, FLT, MAG, MAGERR as rows and the the filter names as columns.
            E.g. Access the magnitude in the z filter with phot_data['z']['MAG'].
        """

        header = self.get_lcs_headers(columns=columns, field=field, \
                                      model=model, base=base, snid=snid, \
                                      limit=limit, sort=sort, shuffle=shuffle, offset=offset, \
                                      big=big, extrasql=extrasql)

        for h in header:
            objid, ptrobs_min, ptrobs_max = h[0:3]
            phot_data = self.get_light_curve(objid, ptrobs_min, ptrobs_max)
            yield h, phot_data




import traceback
import time
import numpy as np
from astrorapid.classify import Classify


# Reset Tensorflow
from tensorflow.keras import backend as K
K.clear_session()


# Load model from disk
classification = Classify(bcut=True, zcut=None)


def delete_indexes(deleteindexes, *args):
    newarrs = []
    for arr in args:
        newarr = np.delete(arr, deleteindexes)
        newarrs.append(newarr)

    return newarrs


def rapid_stage(locus_data):
    locus_properties = locus_data.get_properties()
    objid = locus_properties['alert_id']
    ra = locus_properties['ra']
    dec = locus_properties['dec']
    redshift = 0.  # TODO: Get correct redshift
    mwebv = 0.2  # TODO: Get correct extinction

    # Get redshift from SDSS_gals
    known_redshift = False
    catalog_matches = locus_data.get_astro_object_matches()
    if 'sdss_gals' in catalog_matches:
        redshift = catalog_matches['sdss_gals'][0]['z']
        known_redshift = True

    # Get lightcurve data
    alert_id, mjd, passband, mag, magerr, zeropoint = \
        locus_data.get_time_series(
            'ztf_fid', 'ztf_magpsf', 'ztf_sigmapsf', 'ztf_magzpsci',
            require=['ztf_fid', 'ztf_magpsf', 'ztf_sigmapsf'],
        )

    # Require 2 unique passbands
    if len(np.unique(passband)) < 2:
        print("less than 2 bands")
        return

    # Ignore lightcurves shorter than 3
    if len(mjd) < 3:
        print("less than 3 points")
        return

    # Fill in missing zeropoint values
    zeropoint = np.asarray(zeropoint, dtype=float)
    zpt_median = np.median(zeropoint[(zeropoint != None) & (~np.isnan(zeropoint))])
    zeropoint[zeropoint == None] = zpt_median
    zeropoint[np.isnan(zeropoint)] = zpt_median
    zeropoint = np.asarray(zeropoint, dtype=np.float64)
    if np.any(np.isnan(zeropoint)):
        locus_data.report_error(
            tag='astrorapid_zeropoint_contains_nan',
            data={
                'alert_id': objid,
            },
        )
        return

    # Compute flux
    mag = np.asarray(mag, dtype=np.float64)
    flux = 10. ** (-0.4 * (mag - zeropoint))
    fluxerr = np.abs(flux * magerr * (np.log(10.) / 2.5))

    # Set photflag detections when S/N > 5
    photflag = np.zeros(len(flux))
    photflag[flux / fluxerr > 5] = 4096
    photflag[np.where(mjd == min(mjd[photflag == 4096]))] = 6144

    # Filter out unwanted bands and convert ztf_fid to strings 'g', 'r'
    passband = np.where((passband == 1) | (passband == '1.0'), 'g', passband)
    passband = np.where((passband == 2) | (passband == '2.0'), 'r', passband)
    mjd, passband, flux, fluxerr, zeropoint, photflag = delete_indexes(
        np.where((passband == 3) | (passband == '3.0') | (np.isnan(mag))),
        mjd, passband, flux, fluxerr, zeropoint, photflag
    )

    # Do classification
    light_curves = [
        (mjd, flux, fluxerr, passband, photflag, ra, dec, objid, redshift, mwebv)
    ]
    try:
        classification.known_redshift = known_redshift
        predictions, time_steps = classification.get_predictions(light_curves, return_predictions_at_obstime=True)
    except ValueError:
        locus_data.report_error(
            tag='astrorapid_get_predictions_valueerror',
            data={
                'alert_id': objid,
                'traceback': traceback.format_exc(),
            },
        )
        return

    # Output
    if predictions:
        for i, name in enumerate(classification.class_names):
            # Store properties
            p = predictions[0][-1][i]
            locus_data.set_property('rapid_class_probability_{}'.format(name), p)

            # Send to output streams
            if name == 'Pre-explosion':
                continue
            if p > 0.6:
                stream = 'astrorapid_{}'.format(name.lower().replace('-', '_'))
                locus_data.send_to_stream(stream)



"""
Example usage:

.. code-block:: bash

    python -m astrorapid.read_from_database.read_light_curves_from_snana_fits --offset 0
    --offsetnext 3 --nprocesses 8 --savename 'testing' --combinefiles

"""

import os
import numpy as np
import h5py
import multiprocessing as mp
import pandas as pd
import argparse
import astropy.table as at
import astropy.io.fits as afits
from collections import OrderedDict

from astrorapid.process_light_curves import InputLightCurve


class GetData(object):
    def __init__(self):
        """ Get light curves. """

        self.phot_fields = ['MJD', 'FLT', 'FLUXCAL', 'FLUXCALERR', 'ZEROPT', 'PHOTFLAG']
        self.phot_fields_dtypes = {'FLT': np.str_, 'PHOTFLAG': np.int_}

    def get_light_curve_array(self, phot_file, ptrobs_min, ptrobs_max, standard_zpt=27.5):
        """ Get lightcurve from fits file as an array - avoid some Pandas overhead

        Parameters
        ----------
        objid : str
            The object ID. E.g. objid='DDF_04_NONIa-0004_87287'
        ptrobs_min : int
            Min index of object in _PHOT.FITS.
        ptrobs_max : int
            Max index of object in _PHOT.FITS.

        Return
        -------
        phot_out: pandas DataFrame
            A DataFrame containing the MJD, FLT, FLUXCAL, FLUXCALERR, ZEROPT seperated by each filter.
            E.g. Access the magnitude in the z filter with phot_out['z']['MAG'].
        """

        try:
            phot_HDU = afits.open(phot_file, memmap=True)
        except Exception as e:
            message = f'Could not open photometry file {phot_file}'
            raise RuntimeError(message)

        phot_data = phot_HDU[1].data[ptrobs_min - 1:ptrobs_max]

        phot_dict = OrderedDict()
        filters = list(set(phot_data['FLT']))  # e.g. ['i', 'r', 'Y', 'u', 'g', 'z']
        dtypes = dict(self.phot_fields_dtypes)
        for f in filters:
            fIndexes = np.where(phot_data['FLT'] == f)[0]
            phot_dict[f] = OrderedDict()
            for pfield in self.phot_fields:
                if pfield == 'ZEROPT':
                    phot_dict[f][pfield] = np.repeat(standard_zpt, len(fIndexes))
                elif pfield == 'FLT':
                    true_zpt = phot_data['ZEROPT'][fIndexes]
                    nobs = len(true_zpt)
                    phot_dict[f][pfield] = np.repeat(f.strip(), nobs)
                else:
                    phot_dict[f][pfield] = phot_data[pfield][fIndexes]

                if not pfield in dtypes:
                    dtypes[pfield] = np.float64

        phot_out = pd.DataFrame(phot_dict)
        phot_HDU.close()
        del phot_HDU[1].data
        return phot_out

    def convert_pandas_lc_to_recarray_lc(self, phot, passbands=('u', 'g', 'r', 'i', 'z', 'Y')):
        """
        ANTARES_object not Pandas format broken up by passband
        """
        pbs = passbands
        # name mapping for the defaults in phot_fields
        # any other column is going to become just lowercase it's current name
        name_map = {'FLUXCAL': 'flux', 'FLUXCALERR': 'dflux', 'ZEROPT': 'zpt',\
                        'FLT': 'pb', 'MJD': 'mjd', 'PHOTFLAG': 'photflag'}

        out = None
        out_names = None

        for this_pb in phot:
            # do we know what this passband is
            # this is just a sanity test in case we accidentally retrieve dummy entries with passband = -9
            if this_pb not in pbs:
                continue

            this_pb_lc = phot.get(this_pb)
            if this_pb_lc is None:
                continue

            if out is None:
                out = np.rec.fromarrays(list(this_pb_lc))
                if out_names is None:
                    out_names = list(this_pb_lc.axes[0])
                    out_names = [name_map.get(x, x.lower()) for x in out_names]
            else:
                temp = np.rec.fromarrays(list(this_pb_lc))
                out = np.concatenate((out, temp), axis=-1)

        out.dtype.names = out_names
        return out


def read_light_curves_from_snana_fits_files(save_fname, head_files, phot_files, passbands=('g', 'r'), known_redshift=True):
    """ Save lightcurves from SNANA HEAD AND PHOT FITS files

    Parameters
    ----------
    save_fname : str
        Filename to save hdf5 file.
    head_files : list
        List of SNANA header filepaths (.HEAD files).
    phot_files : list
        List of SNANA photometric filepaths (.PHOT files). Must be in the same order as the phot files.
    passbands : tuple
        passband filters.
    known_redshift : bool
        Whether to use redshift during training.

    """

    getter = GetData()

    store = pd.HDFStore(save_fname)

    for fileidx, headfilepath in enumerate(head_files):
        print(fileidx, headfilepath)
        # Check that phot file correponds to head file
        assert phot_files[fileidx].split('_')[-2] == head_files[fileidx].split('_')[-2]
        header_HDU = afits.open(head_files[fileidx])
        header_data = header_HDU[1].data

        for i, head in enumerate(header_data):
            model_num = head['SIM_TYPE_INDEX']
            snid = head['SNID']
            objid = 'field_{}_base_{}'.format(model_num, snid)
            ptrobs_min = head['PTROBS_MIN']
            ptrobs_max = head['PTROBS_MAX']
            peakmag_g = head['SIM_PEAKMAG_g']
            peakmag_r = head['SIM_PEAKMAG_r']
            redshift = head['SIM_REDSHIFT_HOST']
            dlmu = head['SIM_DLMU']
            peakmjd = head['PEAKMJD']
            mwebv = head['MWEBV']
            mwebv_err = head['MWEBV_ERR']
            ra = head['RA']
            if 'DEC' in header_data.names:
                dec = head['DEC']
            else:
                dec = head['DECL']
            photoz = head['HOSTGAL_PHOTOZ']
            photozerr = head['HOSTGAL_PHOTOZ_ERR']
            print(i, len(header_data))
            phot_data = getter.get_light_curve_array(phot_files[fileidx], ptrobs_min, ptrobs_max)

            lc = getter.convert_pandas_lc_to_recarray_lc(phot_data, passbands=passbands)

            inputlightcurve = InputLightCurve(lc['mjd'], lc['flux'], lc['dflux'], lc['pb'], lc['photflag'], ra,
                                              dec, objid, redshift, mwebv, known_redshift=known_redshift,
                                              training_set_parameters={'class_number': int(model_num), 'peakmjd': peakmjd})

            # TODO: work out why some light curves fail mcmc
            try:
                savepd = inputlightcurve.preprocess_light_curve()
            except Exception as e:
                print("Failed on object", objid, e)
                continue
            store.append(objid, savepd)

    store.close()
    print("saved %s" % save_fname)


def combine_hdf_files(save_dir, combined_savename, training_set_dir):
    fnames = os.listdir(save_dir)
    fname_out = os.path.join(training_set_dir, combined_savename)
    output_file = h5py.File(fname_out, 'w')

    for n, f in enumerate(fnames):
        print(n, f)
        try:
            f_hdf = h5py.File(os.path.join(save_dir, f), 'r')
            for objid in f_hdf.keys():
                objid = objid.encode('utf-8')
                h5py.h5o.copy(f_hdf.id, objid, output_file.id, objid)
            f_hdf.close()
        except OSError as e:
            print("Failed to open file", "f")
            print(e)
    output_file.close()


def create_all_hdf_files(args):
    head_files, phot_files, i, save_dir, passbands, known_redshift = args
    fname = os.path.join(save_dir, 'lc_{}.hdf5'.format(i))
    read_light_curves_from_snana_fits_files(fname, head_files, phot_files, passbands, known_redshift)


def main():
    """
    Save light curves to HDF5 files.
    """

    passbands = ('g', 'r')
    data_release = 'ZTF_20190512'
    field = 'MSIP'
    model = '%'
    known_redshift = True
    dir_name = '/Users/danmuth/PycharmProjects/astrorapid/sim_lc/test_SNIa'

    head_files = []
    phot_files = []
    for subdir, dirs, files in os.walk(dir_name):
        for file in files:
            filepath = os.path.join(subdir, file)
            if filepath.endswith('HEAD.FITS'):
                head_files.append(filepath)
                phot_files.append(filepath.replace('_HEAD.FITS', '_PHOT.FITS'))
            print(filepath)
    # head_files = np.sort(head_files)
    # phot_files = np.sort(phot_files)
    print("Number of head files = {}".format(len(head_files)))

    parser = argparse.ArgumentParser()
    parser.add_argument('-o', "--offset", type=int)
    parser.add_argument('-n', "--offsetnext", type=int)
    parser.add_argument('-m', "--nprocesses", type=int, help='Number of multiprocessing processes. Default is 1.')
    parser.add_argument("--savename", type=str)
    parser.add_argument("--combinefiles", help="Only set this if after this action, all files will have been created.",
                        action='store_true')
    args = parser.parse_args()
    if args.offset is not None:
        offset = args.offset
    else:
        offset = 0
    if args.offsetnext is not None:
        offset_next = args.offsetnext
    else:
        offset_next = 10
    if args.nprocesses is not None:
        nprocesses = args.nprocesses
    else:
        nprocesses = 1
    if args.savename is not None:
        savename = args.savename
    else:
        savename = ""
    print(offset, offset_next)

    training_set_dir = 'training_set_files'
    save_dir = os.path.join(training_set_dir, 'saved_lc_{}_{}_{}'.format(field, data_release, savename))
    if not os.path.exists(save_dir) and offset == 0:
        os.makedirs(save_dir)

    # Multiprocessing
    i_list = np.arange(offset, offset_next)
    print(i_list)
    args_list = []
    file_list = os.listdir(save_dir)
    for i in i_list:
        if 'lc_{}.hdf5'.format(i) not in file_list:
            head_files_part = [head_files[i]]
            phot_files_part = [phot_files[i]]
            print(os.path.join(save_dir, 'lc_{}.hdf5'.format(i)))
            args_list.append((head_files_part, phot_files_part, i, save_dir, passbands, known_redshift))

    if nprocesses == 1:
        for args in args_list:
            create_all_hdf_files(args)
    else:
        pool = mp.Pool(nprocesses)
        results = pool.map_async(create_all_hdf_files, args_list)
        pool.close()
        pool.join()

    if args.combinefiles:
        combine_hdf_files(save_dir, 'saved_lc_{}_{}_{}.hdf5'.format(field, data_release, savename), training_set_dir)


if __name__ == '__main__':
    main()


# keys: dtype=(numpy.record, [('SUBSURVEY', 'S40'), ('SNID', 'S16'), ('IAUC', 'S16'), ('FAKE', '>i2'), ('RA', '>f8'), ('DEC', '>f8'), ('PIXSIZE', '>f4'), ('NXPIX', '>i2'), ('NYPIX', '>i2'), ('CCDNUM', '>i2'), ('SNTYPE', '>i4'), ('NOBS', '>i4'), ('PTROBS_MIN', '>i4'), ('PTROBS_MAX', '>i4'), ('MWEBV', '>f4'), ('MWEBV_ERR', '>f4'), ('REDSHIFT_HELIO', '>f4'), ('REDSHIFT_HELIO_ERR', '>f4'), ('REDSHIFT_FINAL', '>f4'), ('REDSHIFT_FINAL_ERR', '>f4'), ('VPEC', '>f4'), ('VPEC_ERR', '>f4'), ('HOSTGAL_NMATCH', '>i2'), ('HOSTGAL_NMATCH2', '>i2'), ('HOSTGAL_OBJID', '>i8'), ('HOSTGAL_PHOTOZ', '>f4'), ('HOSTGAL_PHOTOZ_ERR', '>f4'), ('HOSTGAL_SPECZ', '>f4'), ('HOSTGAL_SPECZ_ERR', '>f4'), ('HOSTGAL_RA', '>f8'), ('HOSTGAL_DEC', '>f8'), ('HOSTGAL_SNSEP', '>f4'), ('HOSTGAL_DDLR', '>f4'), ('HOSTGAL_CONFUSION', '>f4'), ('HOSTGAL_LOGMASS', '>f4'), ('HOSTGAL_LOGMASS_ERR', '>f4'), ('HOSTGAL_sSFR', '>f4'), ('HOSTGAL_sSFR_ERR', '>f4'), ('HOSTGAL_MAG_g', '>f4'), ('HOSTGAL_MAG_r', '>f4'), ('HOSTGAL_MAGERR_g', '>f4'), ('HOSTGAL_MAGERR_r', '>f4'), ('HOSTGAL_SB_FLUXCAL_g', '>f4'), ('HOSTGAL_SB_FLUXCAL_r', '>f4'), ('PEAKMJD', '>f4'), ('SEARCH_TYPE', '>i4'), ('SIM_MODEL_NAME', 'S32'), ('SIM_MODEL_INDEX', '>i2'), ('SIM_TYPE_INDEX', '>i2'), ('SIM_TYPE_NAME', 'S8'), ('SIM_TEMPLATE_INDEX', '>i4'), ('SIM_LIBID', '>i4'), ('SIM_NGEN_LIBID', '>i4'), ('SIM_NOBS_UNDEFINED', '>i4'), ('SIM_SEARCHEFF_MASK', '>i4'), ('SIM_REDSHIFT_HELIO', '>f4'), ('SIM_REDSHIFT_CMB', '>f4'), ('SIM_REDSHIFT_HOST', '>f4'), ('SIM_REDSHIFT_FLAG', '>i2'), ('SIM_VPEC', '>f4'), ('SIM_DLMU', '>f4'), ('SIM_LENSDMU', '>f4'), ('SIM_RA', '>f8'), ('SIM_DEC', '>f8'), ('SIM_MWEBV', '>f4'), ('SIM_PEAKMJD', '>f4'), ('SIM_MAGSMEAR_COH', '>f4'), ('SIM_AV', '>f4'), ('SIM_RV', '>f4'), ('SIMSED_SALT2x0', '>f4'), ('SIMSED_PARAM(stretch)', '>f4'), ('SIMSED_PARAM(color)', '>f4'), ('SIM_PEAKMAG_g', '>f4'), ('SIM_PEAKMAG_r', '>f4'), ('SIM_EXPOSURE_g', '>f4'), ('SIM_EXPOSURE_r', '>f4'), ('SIM_SUBSAMPLE_INDEX', '>i2')]))



import numpy as np


def find_nearest(array, value):
    """
    Find the index nearest to a given value.
    Adapted from: https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
    """

    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()

    return idx


def calc_luminosity(flux, fluxerr, mu):
    """ Normalise flux light curves with distance modulus.

    Parameters
    ----------
    flux : array
        List of floating point flux values.
    fluxerr : array
        List of floating point flux errors.
    mu : float
        Distance modulus from luminosity distance.

    Returns
    -------
    fluxout : array
        Same shape as input flux.
    fluxerrout : array
        Same shape as input fluxerr.

    """

    d = 10 ** (mu/5 + 1)
    dsquared = d**2

    norm = 1e18

    fluxout = flux * (4 * np.pi * dsquared/norm)
    fluxerrout = fluxerr * (4 * np.pi * dsquared/norm)

    return fluxout, fluxerrout


def get_sntypes():
    sntypes_map = {1: 'SNIa-norm',
                   11: 'SNIa-norm',
                   2: 'SNII',
                   12: 'SNIIpca',
                   14: 'SNIIn',
                   3: 'SNIbc',
                   13: 'SNIbc',
                   5: 'SNIbc',
                   6: 'SNII',
                   41: 'SNIa-91bg',
                   43: 'SNIa-x',
                   45: 'point-Ia',
                   50: 'Kilonova-GW170817',
                   51: 'Kilonova',
                   60: 'SLSN-I',
                   61: 'PISN',
                   62: 'ILOT',
                   63: 'CART',
                   64: 'TDE',
                   70: 'AGN',
                   80: 'RRLyrae',
                   81: 'Mdwarf',
                   83: 'EBE',
                   84: 'Mira',
                   90: 'uLens-BSR',
                   91: 'uLens-1STAR',
                   92: 'uLens-String',
                   93: 'uLens - Point',
                   99: 'Rare'}
    return sntypes_map


def aggregate_sntypes(reverse=False):
    if reverse:
        aggregate_map = {99: (45, 61, 62, 63, 90, 92),
                         1: (11, 1),
                         5: (3,13),
                         6: (2, 12, 14),
                         41: (41,),
                         43: (43,),
                         50: (50,),
                         51: (51,),
                         60: (60,),
                         64: (64,),
                         70: (70,),
                         80: (80,),
                         81: (81,),
                         83: (83,),
                         91: (91,),
                         93: (93,)}
    else:
        aggregate_map = {1: 1,
                         11: 1,
                         2: 6,
                         3: 5,
                         12: 6,
                         13: 5,
                         14: 6,
                         41: 41,
                         43: 43,
                         45: 45,
                         50: 50,
                         51: 51,
                         60: 60,
                         61: 61,
                         62: 62,
                         63: 63,
                         64: 64,
                         70: 70,
                         80: 80,
                         81: 81,
                         83: 83,
                         90: 90,
                         91: 91,
                         92: 92,
                         93: 93,
                         }

    return aggregate_map



import os
import astrorapid.get_training_data
import astrorapid.get_custom_data
from astrorapid.custom_classifier import create_custom_classifier


def main():
    """ Train Neural Network classifier """

    script_dir = os.path.dirname(os.path.abspath(__file__))

    create_custom_classifier(get_data_func=astrorapid.get_training_data.get_data_from_snana_fits,
                             data_dir=os.path.join(script_dir, '..', 'data/ZTF_20190512'),
                             class_nums=(1, 2, 12, 14, 3, 13, 41, 43, 51, 60, 61, 62, 63, 64, 70),
                             reread_data=False,
                             contextual_info=('redshift',),
                             passbands=('g', 'r'),
                             retrain_network=False,
                             train_epochs=50,
                             zcut=0.5,
                             bcut=True,
                             variablescut=True,
                             nprocesses=None,
                             nchunks=10000,
                             aggregate_classes=False,
                             otherchange='',
                             training_set_dir=os.path.join(script_dir, '..', 'training_set_files'),
                             save_dir=os.path.join(script_dir, '..', 'data/saved_light_curves'),
                             fig_dir=os.path.join(script_dir, '..', 'training_set_files', 'Figures', 'ZTF_{}'),
                             plot=True
                             )


if __name__ == '__main__':
    main()



import os
import pickle
import numpy as np
import multiprocessing as mp

from astrorapid.read_snana_fits import read_light_curves_from_snana_fits_files
from astrorapid.process_light_curves import InputLightCurve


def get_custom_data(class_num, data_dir, save_dir, passbands, known_redshift, nprocesses, redo):
    """
    Get data from custom data files.
    You will need to write this function with the following skeleton function:

    def get_data_from_custom(class_num, data_dir, save_dir, passbands, nprocesses, redo):
        # If the data has already been run and processed load it. Otherwise read it and save it
        save_lc_filepath = os.path.join(save_dir, f"lc_classnum_{class_num}.pickle")
        if os.path.exists(save_lc_filepath) and not redo:
            with open(save_lc_filepath, "rb") as fp:  # Unpickling
                light_curves = pickle.load(fp)
        else:
            light_curves = {}
            # Read in data from data_dir and get the mjd, flux, fluxerr, passband, photflag as 1D numpy arrays for
            # each light curve. Get the ra, dec, objid, redshift, mwebv, model_num, peak_mjd as floats or strings.
            # Set whether you'd like to train a model with a known redshift or not. Set known_redshift as a boolean.

            # Enter your own data-reading code here that gets the mjds, fluxes, fluxerrs, passbands, photflags,
            # ras, decs, objids, redshifts, mwebvs, model_nums, peak_mjds for all the light curves from the data_dir

            # Once you have the required data information for each light curve, pass it into InputLightCurve with
            # something like the following code:
            for i, objid in enumerate(objids):
                inputlightcurve = InputLightCurve(mjds[i], fluxes[i], fluxerrs[i], passbands[i], photflags[i],
                                                  ras[i], decs[i], objids[i], redshifts[i], mwebvs[i],
                                                  known_redshift=known_redshift,
                                                  training_set_parameters={'class_number': int(class_num),
                                                                           'peakmjd': peakmjds[i]})
                light_curves[objid] = inputlightcurve.preprocess_light_curve()

            # If you think that reading the data is too slow, you may want to replace the for loop above with
            # multiprocessing. See the example function in get_training_data.py if you need help doing this.

            # Next, we save it:
            with open(save_lc_filepath, "wb") as fp:  # Pickling
                pickle.dump(light_curves, fp)


    Parameters
    ----------
    class_num : int
        Class number. E.g. SNIa is 1. See helpers.py for lookup table.
        E.g. class_num = 1
    data_dir : str
        Directory where data is stored
        E.g. data_dir='data/ZTF_20190512/'
    save_dir : str
        Directory to save processed data
        E.g. save_dir='data/saved_light_curves/'
    passbands : tuple
        Passbands to use.
        E.g. passbands=('g', 'r')
    known_redshift : bool
        Whether to correct the light curves for cosmological time dilation using redshift.
    nprocesses : int or None
        Number of processes to use
    redo : bool
        Whether to redo reading the data and saving the processed data.


    Returns
    -------
    light_curves : dict of astropy.table.Table objects
        e.g light_curves['objid1'] =
            passband   time       flux     fluxErr   photflag
              str1   float32    float32    float32    int32
            -------- -------- ----------- ---------- --------
                   g -46.8942  -48.926975  42.277767        0
                   g -43.9352  -105.35379   72.97575        0
                   g -35.9161  -46.264206    99.9172        0
                   g -28.9377  -28.978344  42.417065        0
                   g -25.9787  109.886566   46.03949        0
                   g -15.0399    -80.2485   80.38155        0
                   g -12.0218    93.51743  113.21529        0
                   g  -6.9585   248.88364 108.606865        0
                   g  -4.0411   341.41498  47.765404        0
                   g      0.0    501.7441   45.37485     6144
                 ...      ...         ...        ...      ...
                   r  40.9147   194.32494  57.836903     4096
                   r  59.9162    67.59185   45.66463     4096
                   r  62.8976    80.85155  44.356197     4096
                   r  65.8974   28.174305   44.75049     4096
                   r  71.8966  -18.790287 108.049774     4096
                   r  74.9297  -3.1707647  125.15057     4096
                   r  77.9341 -11.0205965 125.784676     4096
                   r  80.8576   129.65466   69.99305     4096
                   r  88.8922  -14.259436  52.917866     4096
                   r 103.8734   27.178356 115.537704     4096

    """
    # If the data has already been run and processed load it. Otherwise read it and save it
    save_lc_filepath = os.path.join(save_dir, f"lc_classnum_{class_num}.pickle")
    if os.path.exists(save_lc_filepath) and not redo:
        with open(save_lc_filepath, "rb") as fp:  # Unpickling
            light_curves = pickle.load(fp)
    else:
        light_curves = {}
        # Read in data from data_dir and get the mjd, flux, fluxerr, passband, photflag as 1D numpy arrays for
        # each light curve. Get the ra, dec, objid, redshift, mwebv, model_num, peak_mjd as floats or strings.
        # Set whether you'd like to train a model with a known redshift or not. Set known_redshift as a boolean.

        # Enter your own data-reading code here that gets the mjds, fluxes, fluxerrs, passbands, photflags,
        # ras, decs, objids, redshifts, mwebvs, model_nums, peak_mjds for all the light curves from the data_dir

        # Once you have the required data information for each light curve, pass it into InputLightCurve with
        # something like the following code:
        for i, objid in enumerate(objids):
            inputlightcurve = InputLightCurve(mjds[i], fluxes[i], fluxerrs[i], passbands[i], photflags[i],
                                              ras[i], decs[i], objids[i], redshifts[i], mwebvs[i],
                                              known_redshift=known_redshift,
                                              training_set_parameters={'class_number': int(class_num),
                                                                       'peakmjd': peakmjds[i]})
            light_curves[objid] = inputlightcurve.preprocess_light_curve()

        # If you think that reading the data is too slow, you may want to replace the for loop above with
        # multiprocessing. See the example function in get_training_data.py if you need help doing this.

        # Next, we save it:
        with open(save_lc_filepath, "wb") as fp:  # Pickling
            pickle.dump(light_curves, fp)

    return light_curves



import os
import numpy as np
import h5py
import multiprocessing as mp
import pandas as pd
import astropy.io.fits as afits
from collections import OrderedDict

from astrorapid.process_light_curves import InputLightCurve


class GetData(object):
    def __init__(self):
        """ Get light curves. """

        self.phot_fields = ['MJD', 'FLT', 'FLUXCAL', 'FLUXCALERR', 'ZEROPT', 'PHOTFLAG']
        self.phot_fields_dtypes = {'FLT': np.str_, 'PHOTFLAG': np.int_}

    def get_light_curve_array(self, phot_file, ptrobs_min, ptrobs_max, standard_zpt=27.5):
        """ Get lightcurve from fits file as an array - avoid some Pandas overhead

        Parameters
        ----------
        objid : str
            The object ID. E.g. objid='DDF_04_NONIa-0004_87287'
        ptrobs_min : int
            Min index of object in _PHOT.FITS.
        ptrobs_max : int
            Max index of object in _PHOT.FITS.

        Return
        -------
        phot_out: pandas DataFrame
            A DataFrame containing the MJD, FLT, FLUXCAL, FLUXCALERR, ZEROPT seperated by each filter.
            E.g. Access the magnitude in the z filter with phot_out['z']['MAG'].
        """

        try:
            phot_HDU = afits.open(phot_file, memmap=True)
        except Exception as e:
            message = f'Could not open photometry file {phot_file}'
            raise RuntimeError(message)

        phot_data = phot_HDU[1].data[ptrobs_min - 1:ptrobs_max]

        phot_dict = OrderedDict()
        filters = list(set(phot_data['FLT']))  # e.g. ['i', 'r', 'Y', 'u', 'g', 'z']
        dtypes = dict(self.phot_fields_dtypes)
        for f in filters:
            fIndexes = np.where(phot_data['FLT'] == f)[0]
            phot_dict[f] = OrderedDict()
            for pfield in self.phot_fields:
                if pfield == 'ZEROPT':
                    phot_dict[f][pfield] = np.repeat(standard_zpt, len(fIndexes))
                elif pfield == 'FLT':
                    true_zpt = phot_data['ZEROPT'][fIndexes]
                    nobs = len(true_zpt)
                    phot_dict[f][pfield] = np.repeat(f.strip(), nobs)
                else:
                    phot_dict[f][pfield] = phot_data[pfield][fIndexes]

                if not pfield in dtypes:
                    dtypes[pfield] = np.float64

        phot_out = pd.DataFrame(phot_dict)
        phot_HDU.close()
        del phot_HDU[1].data
        return phot_out

    def convert_pandas_lc_to_recarray_lc(self, phot, passbands=('u', 'g', 'r', 'i', 'z', 'Y')):
        """
        ANTARES_object not Pandas format broken up by passband
        """
        pbs = passbands
        # name mapping for the defaults in phot_fields
        # any other column is going to become just lowercase it's current name
        name_map = {'FLUXCAL': 'flux', 'FLUXCALERR': 'dflux', 'ZEROPT': 'zpt',\
                        'FLT': 'pb', 'MJD': 'mjd', 'PHOTFLAG': 'photflag'}

        out = None
        out_names = None

        for this_pb in phot:
            # do we know what this passband is
            # this is just a sanity test in case we accidentally retrieve dummy entries with passband = -9
            if this_pb not in pbs:
                continue

            this_pb_lc = phot.get(this_pb)
            if this_pb_lc is None:
                continue

            if out is None:
                out = np.rec.fromarrays(list(this_pb_lc))
                if out_names is None:
                    out_names = list(this_pb_lc.axes[0])
                    out_names = [name_map.get(x, x.lower()) for x in out_names]
            else:
                temp = np.rec.fromarrays(list(this_pb_lc))
                out = np.concatenate((out, temp), axis=-1)

        out.dtype.names = out_names
        return out


def read_fits_file(args):
    head_file, phot_file, passbands, known_redshift = args

    getter = GetData()

    light_curves = {}

    header_HDU = afits.open(head_file)
    header_data = header_HDU[1].data

    for i, head in enumerate(header_data):
        class_num = head['SIM_TYPE_INDEX']
        snid = head['SNID']
        objid = '{}_{}'.format(class_num, snid)
        ptrobs_min = head['PTROBS_MIN']
        ptrobs_max = head['PTROBS_MAX']
        redshift = head['SIM_REDSHIFT_HOST']
        peakmjd = head['PEAKMJD']
        mwebv = head['MWEBV']
        ra = head['RA']
        if 'DEC' in header_data.names:
            dec = head['DEC']
        else:
            dec = head['DECL']
        print(i, len(header_data))

        try:
            phot_data = getter.get_light_curve_array(phot_file, ptrobs_min, ptrobs_max)

            lc = getter.convert_pandas_lc_to_recarray_lc(phot_data, passbands=passbands)

            inputlightcurve = InputLightCurve(lc['mjd'], lc['flux'], lc['dflux'], lc['pb'], lc['photflag'], ra,
                                              dec, objid, redshift, mwebv, known_redshift=known_redshift,
                                              training_set_parameters={'class_number': int(class_num), 'peakmjd': peakmjd})
            light_curves[objid] = inputlightcurve.preprocess_light_curve()
        except IndexError as e:
            print("No detections:", e)  # TODO: maybe do better error checking in future
        except AttributeError as e:
            print("phot_data is NoneType", e)  # TODO: maybe fix this later - rare case
        except ValueError as e:
            print("MCMC error while fiiting t0", e)

    return light_curves


def read_light_curves_from_snana_fits_files(head_files, phot_files, passbands=('g', 'r'), known_redshift=True,
                                            nprocesses=1):
    """ Save lightcurves from SNANA HEAD AND PHOT FITS files

    Parameters
    ----------
    head_files : list
        List of filepaths of all SNANA HEAD files.
    phot_files : list
        List of filepaths of all SNANA PHOT files.
    passbands : tuple
        passband filters.
    known_redshift : bool
        Whether to use redshift during training.
    nprocesses : int
        Number of processes to run multiprocessing on.

    """

    args_list = []
    for i, head_file in enumerate(head_files):
        assert phot_files[i].split('_')[-2].split('-')[1] == head_files[i].split('_')[-2].split('-')[1]
        head_files_part = head_files[i]
        phot_files_part = phot_files[i]
        args_list.append((head_files_part, phot_files_part, passbands, known_redshift))

    light_curves = {}
    if nprocesses == 1:
        for args in args_list:
            light_curves.update(read_fits_file(args))
    else:
        pool = mp.Pool(nprocesses)
        results = pool.map_async(read_fits_file, args_list)
        pool.close()
        pool.join()

        outputs = results.get()
        print('combining results...')
        for i, output in enumerate(outputs):
            print(i, len(outputs))
            light_curves.update(output)

    return light_curves

